/* Generated by the Systemtap dtrace wrapper */


#define _SDT_HAS_SEMAPHORES 1


#define STAP_HAS_SEMAPHORES 1 /* deprecated */


#include <sys/sdt.h>

/* USDT_SAMPLE_LIB1_SDT_OPERATION_START_SDT ( uint64_t operation_id, const char* input ) */
#if defined STAP_SDT_V1
#define USDT_SAMPLE_LIB1_SDT_OPERATION_START_SDT_ENABLED() __builtin_expect (operation_start_sdt_semaphore, 0)
#define usdt_sample_lib1_sdt_operation_start_sdt_semaphore operation_start_sdt_semaphore
#else
#define USDT_SAMPLE_LIB1_SDT_OPERATION_START_SDT_ENABLED() __builtin_expect (usdt_sample_lib1_sdt_operation_start_sdt_semaphore, 0)
#endif
__extension__ extern unsigned short usdt_sample_lib1_sdt_operation_start_sdt_semaphore __attribute__ ((unused)) __attribute__ ((section (".probes")));
#define USDT_SAMPLE_LIB1_SDT_OPERATION_START_SDT(arg1, arg2) \
DTRACE_PROBE2 (usdt_sample_lib1_sdt, operation_start_sdt, arg1, arg2)

/* USDT_SAMPLE_LIB1_SDT_OPERATION_END_SDT ( uint64_t operation_id, const char* output ) */
#if defined STAP_SDT_V1
#define USDT_SAMPLE_LIB1_SDT_OPERATION_END_SDT_ENABLED() __builtin_expect (operation_end_sdt_semaphore, 0)
#define usdt_sample_lib1_sdt_operation_end_sdt_semaphore operation_end_sdt_semaphore
#else
#define USDT_SAMPLE_LIB1_SDT_OPERATION_END_SDT_ENABLED() __builtin_expect (usdt_sample_lib1_sdt_operation_end_sdt_semaphore, 0)
#endif
__extension__ extern unsigned short usdt_sample_lib1_sdt_operation_end_sdt_semaphore __attribute__ ((unused)) __attribute__ ((section (".probes")));
#define USDT_SAMPLE_LIB1_SDT_OPERATION_END_SDT(arg1, arg2) \
DTRACE_PROBE2 (usdt_sample_lib1_sdt, operation_end_sdt, arg1, arg2)

