/* SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause) */
/* Copyright (c) 2022 Chen Tao */
#ifndef __JAVAGC_H
#define __JAVAGC_H

struct data_t {
    __u32 cpu;
    __u32 pid;
    __u64 ts;
};

#endif /* __JAVAGC_H */
